/*
	*
	* Copyright 2019 Britanicus <marcusbritanicus@gmail.com>
	*
	* This file is a part of DesQ project (https://gitlab.com/desq/)
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include <unistd.h>

#include <QDir>
#include <QProcess>
#include <QDBusInterface>

#include "DesQSession.hpp"
#include "DesQSessionAdaptor.hpp"

#include <desqwl/DesQWlGlobal.hpp>
#include <desqwl/Registry.hpp>
#include <desqwl/WindowManager.hpp>

DesQSession::DesQSession( DesQSession::SessionType mType ) : QThread() {

	new SessionAdaptor( this );

	QDBusConnection dbusSess = QDBusConnection::sessionBus();
    registered &= dbusSess.registerObject( "/org/DesQ/Session", this );
    registered &= dbusSess.registerService( "org.desq.Session" );

	mSessionType = mType;

	switch( mSessionType ) {
		case DesQSession::DESQ: {
			session = new QSettings( "DesQ", "Session" );
			mSessionName = "DesQ";
			break;
		}

		case DesQSession::OYSTER: {
			session = new QSettings( "Oyster", "Session" );
			mSessionName = "Oyster";
			break;
		}
	}

	/* Power Interface */
	pwr = new Power( this );
};

bool DesQSession::isRunning() {

	return registered;
};

void DesQSession::start( QString name ) {

	mSessionName = name;
	QThread::start();
};

void DesQSession::stop() {

	wl_display *display = DesQWayland::getWlDisplay();

	if ( not display ) {
		qDebug() << "Unable to access wayland display";
		qDebug() << "Session will not be saved.";
	}

	else {
		WaylandRegistry *reg = new WaylandRegistry( display );
		WindowManager *mWM = reg->tlManager();

		QStringList openApps;
		for( WindowHandle *handle: mWM->windowHandles() ) {
			openApps << handle->appID();
			handle->close();
		}
		session->setValue( "Previous/OpenAppDesktops", openApps );
	}

	QProcess proc;
	proc.start( "kill", { "-3", QString::number( wfPID ) } );
	proc.waitForFinished( -1 );

	/* Kill this thread and close the app */
	if ( isRunning() )
		terminate();

	qApp->quit();
};

void DesQSession::run() {

	/* Setup the environmental variables */
	setupEnvironmentVars();

	/* Start the window manager and hence the shell */
	startDesQShell();

	/* Switch on caps lock on login */
	enableNumLock();

	QDBusInterface shell( "org.DesQ.Shell", "/", "org.DesQ.Shell", QDBusConnection::sessionBus() );
	while ( not shell.isValid() ) {
		qDebug() << "Waiting for the shell to be ready...";

		/* Wait for one second, if the shell is not ready */
		sleep( 1 );
	}

	/* Start apps from /etc/xdg/autostart */
	QDir xdgAutoDir( "/etc/xdg/autostart" );
	Q_FOREACH( QString desktop, xdgAutoDir.entryList( QStringList() << "*.desktop" ) ) {

		/* If the same file exists in user's autostart, skip it */
		if ( QFile::exists( QDir::home().filePath( ".config/autostart/" + desktop ) ) )
			continue;

		startDesktop( xdgAutoDir.filePath( desktop ) );
	}

	/* Start apps from ~/.config/autostart */
	QDir homeAutoDir( QDir::home().filePath( ".config/autostart/" ) );
	Q_FOREACH( QString desktop, homeAutoDir.entryList( QStringList() << "*.desktop" ) )
		startDesktop( homeAutoDir.filePath( desktop ) );

	/* Start apps from ~/.config/autostart-scripts */
	QDir scriptAutoDir( QDir::home().filePath( ".config/autostart-scripts/" ) );
	Q_FOREACH( QString script, scriptAutoDir.entryList( QDir::Files ) )
		startScript( scriptAutoDir.filePath( script ) );

	if ( mSessionName.isEmpty() ) {
		mSessionName = session->value( "session" ).toString();
		if ( mSessionName.compare( "New" ) )
			restoreSession( mSessionName );
	}
};

void DesQSession::handleMessages( const QString msg ) {

	if ( msg == "--lock" ) {

		qDebug() << "Trying to lock the active session.";
		QProcess::startDetached( "swaylock", {} );
	}

	else if ( msg == "--logout" ) {

		qDebug() << "Trying to logout from the active session.";
		stop();
	}

	else if ( msg == "--shutdown" ) {

		if ( pwr->systemCanHalt() ) {
			stop();
			pwr->systemHalt();
		}
	}

	else if ( msg == "--reboot" ) {

		if ( pwr->systemCanReboot() ) {
			stop();
			pwr->systemReboot();
		}
	}

	else if ( msg == "--suspend" ) {

		if ( pwr->systemCanSuspend() ) {
			pwr->systemSuspend();
		}
	}

	else if ( msg == "--hibernate" ) {

		if ( pwr->systemCanHibernate() ) {
			pwr->systemHibernate();
		}
	}

	else if ( msg.startsWith( "--session" ) ) {
		#if QT_VERSION >= 0x060000
		QStringList args = msg.split( " ", Qt::SkipEmptyParts );
		#else
		QStringList args = msg.split( " ", QString::SkipEmptyParts );
		#endif

		if ( args.count() == 2 )
			start( args[ 1 ] );

		else if ( args.count() == 1 )
			start();
	}
};

void DesQSession::setupEnvironmentVars() {

	/* Wayland or X11 */
	qputenv( "XDG_SESSION_TYPE", QByteArrayLiteral( "wayland" ) );

	/* Platform for Qt5 */
	qputenv( "QT_QPA_PLATFORM", QByteArrayLiteral( "wayland" ) );

	/* Qt5 theme to be used */
	qputenv( "QT_QPA_PLATFORMTHEME", QByteArrayLiteral( "qt5ct" ) );

	/* Qt5 icon theme to be used */
	qputenv( "QT_QPA_SYSTEM_ICON_THEME", QByteArrayLiteral( "breeze" ) );

	/* Perform Client side decoration: Eventually, we should provide server-side decoration */
	qputenv( "QT_WAYLAND_DISABLE_DECORATION", "0" );

	/* For Mozilla firefox and thunderbird */
	qputenv( "MOZ_ENABLE_WAYLAND", "1" );

	/* Desktop session name */
	qputenv( "XDG_SESSION_DESKTOP", mSessionName.toUtf8() );
	qputenv( "XDG_CURRENT_DESKTOP", mSessionName.toUtf8() );

	/* Desktop session */
	switch( mSessionType ) {
		case SessionType::DESQ: {
			qputenv( "DESKTOP_SESSION", "/usr/share/wayland-sessions/desqsession-wayland" );
			break;
		}

		case SessionType::OYSTER: {
			qputenv( "DESKTOP_SESSION", "/usr/share/wayland-sessions/oystersession" );
			break;
		}
	}

	/* Cursor theme */
	qputenv( "XCURSOR_THEME", "breeze_cursors" );
	qputenv( "XCURSOR_SIZE", "24" );
};

void DesQSession::startDesQShell() {

	qDebug() << "Starting window manager: wayfire";
	QString config = QDir::home().filePath( ".config/DesQ/wayfire.ini" );

	if ( not QFile::exists( config ) ) {
		bool ok = true;
		ok &= QDir::home().mkpath( ".config/DesQ/" );
		ok &= QFile::copy( "/usr/share/desq/wayfire.ini", config );

		/* If we're unable to copy the config file to user's config dir, use the default */
		if ( not ok ) {
			qDebug() << "Error copying wayfire.ini to ~/.config/DesQ/";
			config = "/usr/share/desq/wayfire.ini";
		}
	}

	/* Start wayfire. */
	QProcess::startDetached( "wayfire", { "--config", config }, QDir::homePath(), &wfPID );
};

void DesQSession::startDesktop( QString desktop ) {

	QSettings desktopFile( desktop, QSettings::IniFormat );

	/* If the desktop file is not to be opened in CuboCore, skip it */
	/* Example: cinnamon-settings-daemon-*.desktop should not be opened */
	if ( desktopFile.contains( "Desktop Entry/OnlyShowIn" ) ) {
		QString showOnlyIn = desktopFile.value( "Desktop Entry/OnlyShowIn" ).toString().trimmed();

		/* If @showOnlyIn is not empty and it does not match @mSessionName, don't start the app */
		if ( ( not showOnlyIn.isEmpty() ) and ( showOnlyIn != mSessionName ) )
			return;
	}

	/* If an application is set to be Hidden; do not open it */
	if ( desktopFile.value( "Desktop Entry/Hidden" ).toBool() == true )
		return;

	qDebug() << "Starting desktop" << desktop;

	/* If Exec is valid; execute it */
	QString exec = desktopFile.value( "Desktop Entry/Exec" ).toString();
	if ( not exec.isEmpty() ) {
		#if QT_VERSION >= 0x060000
		QStringList exec2 = exec.trimmed().split( " ", Qt::SkipEmptyParts );
		#else
		QStringList exec2 = exec.trimmed().split( " ", QString::SkipEmptyParts );
		#endif

		if ( QFile::exists( exec2.at( 0 ) ) ) {
			QProcess::startDetached( exec2.takeFirst(), exec2 );
			return;
		}

		else {
			for( QString path : qgetenv( "PATH" ).split( ':' ) ) {
				if ( QFile::exists( path + "/" + exec2.at( 0 ) ) ) {
					QProcess::startDetached( exec2.takeFirst(), exec2 );
					return;
				}
			}
		}

		return;
	}
};

void DesQSession::startScript( QString script ) {

	/* @script can be a link, get the real path */
	QString realScript = QFileInfo( script ).canonicalFilePath();
	qDebug() << "Starting script" << script;
	QProcess::startDetached( realScript, {} );
};

void DesQSession::enableNumLock() {

	qDebug() << "Enabling NumLock: This feature is not implemented as of now.";
};

void DesQSession::restoreSession( QString name ) {

	for( QString desktop: session->value( name + "/OpenAppDesktops" ).toStringList() ) {
		startDesktop( desktop );
	}
};
